from typing import Optional, Callable, Union
from datetime import datetime
import colortk
from time import time
from .objects import LogManager, TimeLogConfig, Colorizer


def __logger_name(__file: str):
    result = __file.replace("/", ".").removesuffix(".py")
    if ".." in result:
        result = result.split("..", 1)[1]
    while result.startswith("."):
        result = result.removeprefix(".")
    return result


log_manager = LogManager(
    colorizer=Colorizer(
        name="log",
        colorize_base=colortk.colorizer("cyan")
    )
)


def __form_format_result(config:TimeLogConfig) -> Callable[[object], str]:
    if config.include_result:
        format_result = config.format_result
        message_limit = config.message_limit
        format_base = lambda __result: str(format_result(__result))
        if message_limit:
            format_limit = lambda __result: format_base(__result)[:message_limit]
            format_unline = lambda __result: format_limit(__result).replace("\n", " ")
        else:
            format_unline = lambda __result:format_base(__result).replace("\n"," ")

        result = lambda __result:f": [{format_unline(__result)}]"
    else:
        result = lambda __result:""
    return result


UTC_TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f UTC'
def __form_timestamp_generator(config:TimeLogConfig):
    timestamp_strformat = None
    if config.timestamp_strformat:
        timestamp_strformat = config.timestamp_strformat
    elif config.include_timestamp:
        timestamp_strformat = UTC_TIMESTAMP_FORMAT
    if timestamp_strformat is not None:
        def generate_timestamp():
            return datetime.utcnow().strftime(timestamp_strformat)
    else:
        def generate_timestamp():
            return ""
    return generate_timestamp
        


def __form_log_message(config: TimeLogConfig, func):
    name_func = func.__name__
    colorize_time = colortk.colorizer("green")
    sublogger_name = __logger_name(config.file)
    colorizer_module = log_manager.sub_colorizer(sublogger_name, color=config.color)
    module_id = colorizer_module.colorized_id()

    colorizer_func = log_manager[sublogger_name].sub_colorizer(name_func)
    log_manager[sublogger_name][name_func] = colorizer_func
    func_id = colorizer_func.colorized_id()

    form_result = __form_format_result(config)
    timestamp_generator = __form_timestamp_generator(config)

    def form_time(__elapsed:str):
        m_time_base = f"( {str(__elapsed)[:12]}s )"
        return colorize_time(m_time_base)

    def log_message(result, elapsed: str):
        args = (f"INFO:     {timestamp_generator()}",form_time(elapsed), ": ", module_id, ".", func_id,form_result(result))
        print(*args, sep="")

    return log_message


def __timelog_sync(config: TimeLogConfig):
    def sublog(func):
        log_message = __form_log_message(config=config, func=func)

        def inner(*args, **kwargs):
            start = time()
            result = func(*args, **kwargs)
            elapsed = time() - start
            log_message(result=result, elapsed=elapsed)
            return result

        return inner

    return sublog


def __timelog_async(config: TimeLogConfig):
    def sublog(func):
        log_message = __form_log_message(config=config, func=func)

        async def inner(*args, **kwargs):
            start = time()
            result = await func(*args, **kwargs)
            elapsed = time() - start
            log_message(result=result, elapsed=elapsed)
            return result

        return inner

    return sublog


def timelog(
    file: str,
    color: Optional[Union[str, int]] = None,
    include_result: bool = True,
    format_result: Callable[[object], str] = lambda value: value,
    message_limit: Optional[int] = None,
    include_timestamp:bool = False,
    timestamp_strformat:Optional[str]=None,
    is_async: bool = False,
):
    config_kwargs = dict(
        file=file,
        color=color,
        message_limit=message_limit,
        format_result=format_result,
        include_result=include_result,
        timestamp_strformat=timestamp_strformat,
        include_timestamp=include_timestamp
    )
    config = TimeLogConfig(**config_kwargs)
    return (__timelog_async if is_async else __timelog_sync)(config=config)



import logging
from dataclasses import dataclass, field
from typing import Callable, List
from loguru import logger


@dataclass
class LoggerConfig:

    startswith: List[str] = field(default_factory=list)
    endswith: List[str] = field(default_factory=list)
    contains: List[str] = field(default_factory=list)
    include: List[str] = field(default_factory=list)
    exclude: List[str] = field(default_factory=list)

    @property
    def is_empty(self) -> bool:
        return not (
            self.startswith
            or self.endswith
            or self.contains
            or self.exclude
            or self.include
        )

    def __contains__(self, __name: str):
        result = False
        for base in self.startswith:
            if __name.startswith(base):
                result = True
                break
        else:
            for base in self.endswith:
                if __name.endswith(base):
                    result = True
                    break
            else:
                for base in self.contains:
                    if base in __name:
                        result = True
                        break
        if result:
            if __name in self.exclude:
                result = False
        elif __name in self.include:
            result = True
        return result


class InterceptHandler(logging.Handler):
    def __init___(self, record_callback: Callable[[logging.LogRecord], str]):
        super().__init__()
        self.record_callback = record_callback

    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        message = self.record_callback(record)
        logger.opt(depth=depth, exception=record.exc_info).log(level, message)



import logging
import sys
from pprint import pformat
from typing import Any
from loguru import logger
from loguru._defaults import LOGURU_FORMAT
from .objects import InterceptHandler, LoggerConfig


def format_record(record: dict) -> str:

    format_string = LOGURU_FORMAT
    if record["extra"].get("payload") is not None:
        record["extra"]["payload"] = pformat(
            record["extra"]["payload"], indent=4, compact=True, width=88
        )
        format_string += "\n<level>{extra[payload]}</level>"

    format_string += "{exception}\n"
    return format_string


def init_logging(
    logger_config: LoggerConfig = LoggerConfig(
        startswith=[], endswith=[], contains=[], include=[], exclude=[]
    )
):

    if not logger_config.is_empty:
        system_loggers = [
            logging.getLogger(name) for name in logging.root.manager.loggerDict
        ]
        system_loggers_names = [sublogger.name for sublogger in system_loggers]
        system_loggers_names.sort()
        logged_modules = [
            sublogger for sublogger in system_loggers if sublogger.name in logger_config
        ]
        for sublogger in logged_modules:
            sublogger.handlers = []

        intercept_handler = InterceptHandler(level=logging.DEBUG)
        for logged_module in logged_modules:
            logging.getLogger(logged_module.name).handlers = [intercept_handler]

        logger.configure(
            handlers=[
                {"sink": sys.stdout, "level": logging.DEBUG, "format": format_record}
            ]
        )


def log(__level: int or str, __message: str, *args: Any, **kwargs: Any):
    logger.log(__level, __message, *args, **kwargs)


def info(__message: str, *args: Any, **kwargs: Any):
    logger.info(__message, *args, **kwargs)


def error(__message: str, *args: Any, **kwargs: Any):
    logger.error(__message, *args, **kwargs)



import colortk
from dataclasses import dataclass, field
from typing import Optional, Callable, Dict, Union, Any


@dataclass(frozen=False, order=True)
class Colorizer:

    name: str
    colorize_base: Callable[[str], str]
    colorizer_map: Dict[str, Any] = field(default_factory=dict)

    def __getitem__(self, __key: str):
        result: Colorizer = self.colorizer_map[__key]
        return result

    def __contains__(self, __key: str):
        return __key in self.colorizer_map

    def __setitem__(self, __key: str, value):
        self.colorizer_map[__key] = value

    def colorized_id(self):
        return self.colorize(self.name)

    def colorize(self, __str: str):
        return self.colorize_base(__str)

    def sub_colorizer(self, __key: str, color: Optional[Union[str, int]] = None):
        if __key in self:
            colorizer = self[__key]
        else:
            name = __key
            colorize_base = colortk.colorizer(color=color)
            colorizer = Colorizer(name=name, colorize_base=colorize_base)
            self[__key] = colorizer
        return colorizer


@dataclass(frozen=False, order=True)
class LogManager:

    colorizer: Colorizer

    def __getitem__(self, key: str):
        return self.colorizer[key]

    def __setitem__(self, key: str, value):
        self.colorizer[key] = value

    def sub_colorizer(self, __key: str, color: Optional[Union[str, int]] = None):
        if __key in self.colorizer:
            result = self.colorizer[__key]
        else:
            colorizer = self.colorizer
            result = colorizer.sub_colorizer(__key, color=color)
            self.colorizer = colorizer
        return result


@dataclass(frozen=False, order=True)
class TimeLogConfig:
    file: str
    color: Optional[Union[str, int]]
    include_result: bool
    format_result: Callable[[object], str]
    message_limit: int
    include_result: bool
    include_timestamp: bool
    timestamp_strformat:Optional[str]
